(* ::Package:: *)

(* ::Title:: *)
(*Twitter*)


(* ::Author:: *)
(*Rob Raguet-Schofield*)
(*ragfield@gmail.com*)


(* ::Section:: *)
(*Initialization*)


BeginPackage["Twitter`", {"JLink`"}];


TwitterSession::usage = "TwitterSession[...] is an object that contains login credentials for a Twitter account.";
TwitterSessionScreenName::usage = "TwitterSessionScreenName[session] returns the Twitter account name associated with the given session.";
TwitterSessionUser::usage = "TwitterSessionUser[session] returns the TwitterUser object associated with the given session.";

TwitterSessionOpen::usage = "TwitterSessionOpen[(\"User\"->\"screen_name\", (\"Password\"->\"password\"))] creates a new TwitterSession with the given account credentials, or (if no credentials are specified) prompts the user for a user name and password.";
TwitterSessionClose::usage = "TwitterSessionClose[session] terminates the TwitterSession and releases the resources associated with it.";
TwitterSessionAuthorize::usage = "TwitterSessionAuthorize[session] can be used to reauthorize a TwitterSession with new account credentials.";
TwitterSessionAuthorizedQ::usage = "TwitterSessionAuthorizedQ[session] returns True if the given TwitterSession is currently authorized, or False if it is not currently authorized.";

TwitterUser::usage = "TwitterUser[...] is an object that contains information about a specific Twitter user account.";
TwitterUserID::usage = "TwitterUserID[user] returns the unique integer identifier for the given TwitterUser.";
TwitterUserName::usage = "TwitterUserName[user] returns the full name of the given TwitterUser (if available).";
TwitterUserScreenName::usage = "TwitterUserScreenName[user] returns the screen name of the given TwitterUser.";
TwitterUserLocation::usage = "TwitterUserLocation[user] returns the Twitter profile location of the given TwitterUser (if available).";
TwitterUserHomePage::usage = "TwitterUserHomePage[user] returns the Twitter profile home page URL of the given TwitterUser (if available).";
TwitterUserDescription::usage = "TwitterUserDescription[user] returns the Twitter profile bio information of the given TwitterUser (if available).";
TwitterUserProfilePage::usage = "TwitterUserProfilePage[user] returns the URL of the Twitter web page of the given TwitterUser.";
TwitterUserProfileImage::usage = "TwitterProfileImage[user] returns the profile image of the given TwitterUser.";
TwitterUserStatusCount::usage = "TwitterUserStatusCount[user] returns the number of status messages posted by the given user.";
TwitterUserFriendCount::usage = "TwitterUserFriendCount[user] returns the number of friends the given user has.";
TwitterUserFollowerCount::usage = "TwitterUserFollowerCount[user] returns the number of followers the given user has.";
TwitterUserFavoriteCount::usage = "TwitterUserFavoriteCount[user] returns the number of favorite status messages the given user has.";
TwitterUserProtectedQ::usage = "TwitterUserProtectedQ[user] returns true if the given user has protected their updates, False if they have not.";
TwitterUserDate::usage = "TwitterUserDate[user] returns the account creation date of the given user.";
TwitterUserInfo::usage = "TwitterUserInfo[user] returns a styled text representation of the name, location, home page, description of the give TwitterUser.";
TwitterUserUI::usage = "TwitterUserUI[user] returns a graphical user interface element for the given TwitterUser.";


TwitterStatus::usage = "TwitterStatus[...] is an object that contains information about a specific Twitter status message.";
TwitterStatusID::usage = "TwitterStatusID[status] returns the unique integer identifier for the given TwitterStatus.";
TwitterStatusUser::usage = "TwitterStatusUser[status] returns the TwitterUser who posted the given TwitterStatus.";
TwitterStatusText::usage = "TwitterStatusText[status] returns the textual contents of the given TwitterStatus.";
TwitterStatusDate::usage = "TwitterStatusDate[status] returns the posted date of the given TwitterStatus.";
TwitterStatusPage::usage = "TwitterStatusPage[status] returns the Twitter URL of the given TwitterStatus.";
TwitterStatusSource::usage = "TwitterStatusSource[status] returns text or hyperlink representation of the application used to post the given TwitterStatus.";
TwitterStatusReplyID::usage = "TwitterStatusReplyID[status] returns the unique integer identifier of the TwitterStatus to which the given TwitterStatus is a reply, or 0 if the given TwitterStatus is not a reply.";
TwitterStatusReplyUserID::usage = "TwitterStatusReplyUserID[status] returns the unique integer identifier of the TwitterUser to which the given TwitterStatus replies, or 0 if the given TwitterStatus is not a reply.";
TwitterStatusUI::usage = "TwitterStatusUI[status] returns a graphical user interface element for the given TwitterStatus.";


TwitterSetStatus::usage = "TwitterSetStatus[session, expr] sets the status of the user associated with session to the string representation of expr.";
TwitterReply::usage = "TwitterReply[session, status|id, expr] replies to the given status with the string representation of expr.";
TwitterDeleteStatus::usage = "TwitterDeleteStatus[session, status|id] deletes the status specified by the given TwitterStatus or unique integer identifier.";


TwitterPublicTimeline::usage = "TwitterPublicTimeline[(\"Page\"->n)] returns the nth page of most recent public Twitter status messages.";
TwitterFriendsTimeline::usage = "TwitterFriendsTimeline[session (, \"Page\"->n)] returns the nth page of most recent Twitter status messages of friends of the user associated with session.";
TwitterUserTimeline::usage = "TwitterUserTimeline[session|user|id (, \"Page\"->n)] returns the nth page of most recent Twitter status messages of the specified user.";
TwitterReplies::usage = "TwitterReplies[session (, \"Page\"->n)] returns the nth page of most recent replies to the user associated with session.";


TwitterFriends::usage = "TwitterFriends[session (, user|id (, \"Page\"->n))] returns the nth page of users who are friends of the given user.";
TwitterAllFriends::usage = "TwitterAllFriends[session (, user|id)] returns a list of all users who are friends of the given user.";
TwitterFollowers::usage = "TwitterFollowers[session (, user|id (, \"Page\"->n))] returns the nth page of users who are followers of the given user.";
TwitterAllFollowers::usage = "TwitterAllFollowers[session (, user|id)] returns a list of all users who are followers of the given user.";


HTTP::err = "Unknown Error (`1`) for URL `2`";
HTTP::e401 = "Not Authorized for URL `1`";
HTTP::e403 = "Forbidden for URL `1`";
HTTP::e404 = "Not Found for URL `1`";
HTTP::e500 = "Internal Server Error for URL `1`";
HTTP::e502 = "Bad Gateway (server is down) for URL `1`";
HTTP::e503 = "Service Unavailable (server is busy) for URL `1`";


HTTPClient::usage = "HTTPClient is the object returned by HTTPOpen.";


HTTPOpen::usage = "HTTPOpen[] returns a new HTTPClient object.";


HTTPClose::usage = "HTTPClose[client] frees resources associated with client when it is no longer needed.";


HTTPAuthorize::usage = "";


HTTPGet::usage = "";
HTTPPost::usage = "";


HTTPDateString::usage = "";
HTTPDateList::usage = "";


Begin["`Private`"];


(* ::Section:: *)
(*Implementation*)


(* ::Subsection::Closed:: *)
(*HTTP.m*)


LoadJavaClass["org.apache.commons.httpclient.auth.AuthScope"];


Format[client_HTTPClient, StandardForm] := "HTTPClient[<...>]";


httpClient[client_HTTPClient] := First[client];


HTTPOpen[] := HTTPClient[
	JavaNew["org.apache.commons.httpclient.HttpClient"]
];


HTTPClose[client_HTTPClient] := ReleaseJavaObject[httpClient[client]];


(* ::Input:: *)
(*client=HTTPOpen[]*)


(* ::Input:: *)
(*HTTPClose[client]*)


Options[HTTPAuthorize] = {
	"User" -> "",
	"Password" -> "",
	"Server" -> AuthScope`ANYUHOST,
	"Port" -> AuthScope`ANYUPORT
};


HTTPAuthorize[client_HTTPClient, opts___?OptionQ] := Module[
	{user, pass, res = True, creds, scope, server, port},
	user = "User" /. {opts} /. Options[HTTPAuthorize];
	pass = "Password" /. {opts} /. Options[HTTPAuthorize];
	server = "Server" /. {opts} /. Options[HTTPAuthorize];
	port = "Port" /. {opts} /. Options[HTTPAuthorize];
	
	If[StringLength[user] === 0 || StringLength[pass] === 0,
		DynamicModule[{dynUser = user, dynPass = pass},
			res = ChoiceDialog[Column[{
				TextCell["Login to " <> server <> ":" <> ToString[port],
					FontSize->CurrentValue[{"ControlsFontSize", Large}]],
				TextCell["User name:"],
				InputField[Dynamic[dynUser], String],
				TextCell["Password:"],
				InputField[Dynamic[dynPass], String,
					BaseStyle->{FontOpacity->0}]
			}]];
			user = dynUser;
			pass = dynPass;
		]
	];
	If[!res, pass = ""; Return[$Canceled]];
	JavaBlock[
		creds = JavaNew[
			"org.apache.commons.httpclient.UsernamePasswordCredentials",
			user, pass];
		scope = JavaNew[
			"org.apache.commons.httpclient.auth.AuthScope",
			server, port, AuthScope`ANYUREALM];
		httpClient[client]@getState[]@setCredentials[scope, creds];
		pass = "" (* Null out module variable *)
	];
	True
];


(* ::Input:: *)
(*HTTPAuthorize[client,"User"->"ragfield","Server"->"twitter.com","Port"->80]*)


(* ::Input:: *)
(*httpClient[client]@getState[]@getCredentials[]*)


HTTPExecute[client_HTTPClient, method_?JavaObjectQ, url_String] := Module[
	{res},
	res = httpClient[client]@executeMethod[method];
	Switch[res,
		200, Return@JavaBlock[
			method@getResponseBodyAsString[]
		],
		401, Message[HTTP::e401, url],
		403, Message[HTTP::e403, url],
		404, Message[HTTP::e404, url],
		500, Message[HTTP::e500, url],
		502, Message[HTTP::e502, url],
		503, Message[HTTP::e503, url],
		_, Message[HTTP::err, res, url]
	];
	$Failed
];


HTTPGet[client_HTTPClient, url_String] := Module[
	{method, res},
	method = JavaNew[
		"org.apache.commons.httpclient.methods.GetMethod",
		url];
	res = HTTPExecute[client, method, url];
	ReleaseJavaObject[method];
	res
];


(* ::Input:: *)
(*HTTPGet[client,"http://twitter.com/account/verify_credentials.xml"]*)


(* ::Input:: *)
(*HTTPGet[client, "http://google.com/doesnotexist"]*)


HTTPPost[client_HTTPClient, url_String] := Module[
	{method, res},
	method = JavaNew[
		"org.apache.commons.httpclient.methods.PostMethod",
		url];
	res = HTTPExecute[client, method, url];
	ReleaseJavaObject[method];
	res
];


(* ::Text:: *)
(*Sun, 06 Nov 1994 08:49:37 GMT  ; RFC 822, updated by RFC 1123*)


httpDateString[gmt_List] :=
	DateString[gmt, {
		"DayNameShort", ", ",
		"Day", " ",
		"MonthNameShort", " ",
		"Year", " ",
		"Time", " GMT"
	}];


(* ::Input:: *)
(*httpDateString[Date[0]]*)


HTTPDateString[] := httpDateString[Date[0]];


(* ::Input:: *)
(*HTTPDateString[]*)


HTTPDateString[{
	year_Integer, mon_Integer, day_Integer,
	hour_Integer, min_Integer, sec_?NumericQ
}] := httpDateString[{year, mon, day, hour, min, sec}];


(* ::Input:: *)
(*HTTPDateString[Date[]]*)


(* ::Input:: *)
(*HTTPDateString[Date[0]]*)


HTTPDateString[time_?NumericQ] :=
	HTTPDateString[ToDate[time]];


(* ::Input:: *)
(*HTTPDateString[AbsoluteTime[]]*)


(* ::Input:: *)
(*HTTPDateString[AbsoluteTime[0]]*)


pmonth = Alternatives["Jan", "Feb", "Mar", "Apr", "May", "Jun",
	"Jul", "Aug", "Sep", "Oct", "Nov", "Dec"];


pweekday = Alternatives["Monday", "Tuesday", "Wednesday",
	"Thursday", "Friday", "Saturday", "Sunday"];


pwkday = Alternatives["Mon", "Tue", "Wed", "Thu", "Fri", "Sat", "Sun"];


ptwodigit = DigitCharacter ~~ DigitCharacter;


pfourdigit = ptwodigit ~~ ptwodigit;


ptime = ptwodigit ~~ ":" ~~ ptwodigit ~~ ":" ~~ ptwodigit;


pdate1 = ptwodigit ~~ " " ~~ pmonth ~~ " " ~~ pfourdigit;


pdate2 = ptwodigit ~~ "-" ~~ pmonth ~~ "-" ~~ ptwodigit;


pdate3 = pmonth ~~ " " ~~ Alternatives[ptwodigit, " " ~~ DigitCharacter];


rfc1123 = pwkday ~~ ", " ~~ pdate1 ~~ " " ~~ ptime ~~ " GMT";


rfc850 = pweekday ~~ ", " ~~ pdate2 ~~ " " ~~ ptime ~~ " GMT";


asctime = pwkday ~~ " " ~~ pdate3 ~~ " " ~~ ptime ~~ " " ~~ pfourdigit;


httpdate = Alternatives[rfc1123, rfc850, asctime];


(* ::Input:: *)
(*str = HTTPDateString[]*)


HTTPDateList[str_String] :=
	(* normalize the date to remove the +0000 time zone because DateList doesn't understand it *)
	DateList@StringReplace[str, RegularExpression[
		"(Mon|Tue|Wed|Thu|Fri|Sat|Sun),?\\s(Jan|Feb|Mar|Apr|May|Jun|Jul|Aug|Sep|Oct|Nov|Dec)\\s(\\d\\d)\\s(\\d\\d):(\\d\\d):(\\d\\d)\\s([+-]\\d*\\s)?(\\d*)"]:>"$2 $3 $8 $4:$5:$6"
	];


(* ::Input:: *)
(*HTTPDateList["Mon Feb 04 16:11:54 +0000 2008"]*)


(* ::Subsection:: *)
(*Utility Methods*)


hexEncode[s_String] := StringJoin[
	Riffle[IntegerString[ToCharacterCode[s, "UTF-8"], 16, 2], "%", {1, -2, 2}]
]


getQueryString[args_List] :=
	StringJoin@@Riffle[StringJoin[
		ToString@First@#, "=", ToString@Last@#]& /@ args, "&"]


getURLWithQueryString[url_String, args_List] := Module[
	{query = getQueryString[args]},
	If[StringLength[query] =!= 0,
		url <> "?" <> query,
		url
	]
];


utcToLocal[{year_, mon_, day_, h_, m_, s_}] := JavaBlock@Module[
	{utcZone, localZone, utcCal, localCal, date},
	LoadJavaClass["java.util.TimeZone"];
	LoadJavaClass["java.util.GregorianCalendar"];
	utcZone = TimeZone`getTimeZone["UTC"];
	localZone = TimeZone`getDefault[];
	utcCal = JavaNew["java.util.GregorianCalendar", utcZone];
	localCal = JavaNew["java.util.GregorianCalendar", localZone];
	utcCal@set[year, mon - 1, day, h, m, Round[s]];
	localCal@setTime[utcCal@getTime[]];
	(localCal@get[#]& /@ {
		Calendar`YEAR, Calendar`MONTH, Calendar`DATE,
		Calendar`HOURUOFUDAY, Calendar`MINUTE, Calendar`SECOND
	}) + {0, 1, 0, 0, 0, 0.} (* convert seconds to Real *)
];


xmlElementRules[str_String] := str;


xmlElementRules[XMLElement[tag_String, _, {val_String}]] :=
	Rule[tag, val];


xmlElementRules[XMLElement[tag_String, _, {}]] :=
	Rule[tag, ""];


xmlElementRules[XMLElement[tag_String, attrs_List, elems:{___, _XMLElement, ___}]] :=
	Rule[tag, xmlElementRules /@ elems];


(* ::Subsection:: *)
(*TwitterSession*)


Format[twitter_TwitterSession, StandardForm] := StringForm[
	"TwitterSession[<`1`>]",
	TwitterSessionScreenName[twitter]
];


TwitterSessionHTTPClient[twitter_TwitterSession] := Part[twitter, 1];


TwitterSessionScreenName[twitter_TwitterSession] := Part[twitter, 2];


TwitterSessionUser[twitter_TwitterSession] := Part[twitter, 3];


TwitterSessionOpen[opts___?OptionQ] := Module[
	{user, twitter, creds},
	user = "User" /. {opts} /. {"User"->""};
	twitter = TwitterSession[HTTPOpen[], user];
	creds = TwitterSessionAuthorize[twitter];
	If[Head[creds] =!= TwitterUser,
		HTTPClose[TwitterSessionHTTPClient[twitter]];
		Return[$Failed]
	];
	If[StringLength[user] === 0, twitter[[2]] = TwitterUserScreenName[creds]];
	AppendTo[twitter, creds]
];


TwitterSessionClose[twitter_TwitterSession] :=
	HTTPClose[TwitterSessionHTTPClient[twitter]];


(* ::Input:: *)
(*twitter=TwitterSessionOpen["User"->"ragfield"]*)


(* ::Input:: *)
(*TwitterSessionClose[twitter]*)


TwitterSessionAuthorize[twitter_TwitterSession, opts___?OptionQ] := Module[
	{},
	If[HTTPAuthorize[TwitterSessionHTTPClient[twitter],
		"Server"->"twitter.com", "Port"->443, 
		"User"->TwitterSessionScreenName[twitter], opts] =!= True,
			Return[$Canceled]
	];
	TwitterSessionCredentials[twitter]
];


(* ::Input:: *)
(*TwitterSessionAuthorize[twitter]*)


TwitterSessionCredentials[twitter_TwitterSession] := Module[
	{res},
	res = Quiet[
		HTTPGet[TwitterSessionHTTPClient[twitter],
			"https://twitter.com/account/verify_credentials.xml"],
		{HTTP::e401}
	];
	If[res === $Failed, res];
	
	res = Quiet@ImportString[res, "XML"];
	If[res === $Failed, res];
	
	TwitterUser[res[[2]]]
];



(* ::Input:: *)
(*TwitterSessionCredentials[twitter]*)


TwitterSessionAuthorizedQ[twitter_TwitterSession] := Module[
	{user = TwitterSessionCredentials[twitter]},
	(TwitterSessionScreenName[twitter] == TwitterUserScreenName[user])
];


(* ::Input:: *)
(*TwitterSessionAuthorizedQ[twitter]*)


(* ::Input:: *)
(*Import["https://twitter.com/account/verify_credentials.xml"]*)


(* ::Subsection:: *)
(*TwitterUser*)


Format[user_TwitterUser, StandardForm] := StringForm[
	"TwitterUser[<`1`>]",
	TwitterUserScreenName[user]
];


TwitterUser[XMLElement["user", _, elems_List]] := Module[
	{rules=  xmlElementRules /@ elems},
	rules = rules /. {
		Rule[x:("id"|"statuses_count"|
			"friends_count"|"followers_count"|
			"favorites_count"), y_] :> Rule[x, ToExpression[y]],
		Rule[x:"protected", "true"] :> Rule[x, True],
		Rule[x:"protected", "false"] :> Rule[x, False],
		Rule["created_at", y_] :>
			Rule["created_at", utcToLocal[HTTPDateList[y]]]
	};
	TwitterUser@@rules
];


(* ::Input:: *)
(*user=TwitterSessionUser[twitter]*)


TwitterUser[id_Integer] := Module[
	{xml},
	xml = Import["https://twitter.com/users/show.xml?user_id=" <>
		ToString[id]];
	If[xml === $Failed, Return[$Failed]];
	TwitterUser@First@Cases[xml, XMLElement["user", _, _], \[Infinity]]
];


(* ::Input:: *)
(*TwitterUser[12920162]*)


TwitterUser[screenName_String] := Module[
	{xml},
	xml = Import["https://twitter.com/users/show.xml?screen_name=" <>
		screenName, "XML"];
	If[xml === $Failed, Return[$Failed]];
	TwitterUser@First@Cases[xml, XMLElement["user", _, _], \[Infinity]]
];


(* ::Input:: *)
(*TwitterUser["ragfield"]*)


TwitterUserID[user_TwitterUser] := (
	"id" /. List@@user /. {"id"->0}
);


(* ::Input:: *)
(*TwitterUserID[user]*)


TwitterUserName[user_TwitterUser] := (
	"name" /. List@@user /. {"name"->""}
);


(* ::Input:: *)
(*TwitterUserName[user]//InputForm*)


TwitterUserScreenName[user_TwitterUser] := (
	"screen_name" /. List@@user /. {"screen_name"->""}
);


(* ::Input:: *)
(*TwitterUserScreenName[user]*)


TwitterUserLocation[user_TwitterUser] := (
	"location" /. List@@user /. {"location"->""}
);


(* ::Input:: *)
(*TwitterUserLocation[user]*)


TwitterUserHomePage[user_TwitterUser] := (
	"url" /. List@@user /. {"url"->""}
);


(* ::Input:: *)
(*TwitterUserHomePage[user]*)


TwitterUserDescription[user_TwitterUser] := (
	"description" /. List@@user /. {"description"->""}
);


(* ::Input:: *)
(*TwitterUserDescription[user]//InputForm*)


TwitterUserProfilePage[user_TwitterUser] := StringJoin[
	"http://twitter.com/",
	TwitterUserScreenName[user]
];


(* ::Input:: *)
(*TwitterUserProfilePage[user]*)


TwitterUserProfileImage[user_TwitterUser] := Module[
	{img},
	img = "profile_image_url" /. List@@user /. {
		"profile_image_url"->"http://static.twitter.com/images/default_profile_normal.png"
	};
	Import[img]
];


(* ::Input:: *)
(*TwitterUserProfileImage[user]*)


TwitterUserStatusCount[user_TwitterUser] := (
	"statuses_count" /. List@@user /. {"statuses_count"->0}
);


(* ::Input:: *)
(*TwitterUserStatusCount[user]*)


TwitterUserFriendCount[user_TwitterUser] := (
	"friends_count" /. List@@user /. {"friends_count"->0}
);


(* ::Input:: *)
(*TwitterUserFriendCount[user]*)


TwitterUserFollowerCount[user_TwitterUser] := (
	"followers_count" /. List@@user /. {"followers_count"->0}
);


(* ::Input:: *)
(*TwitterUserFollowerCount[user]*)


TwitterUserFavoriteCount[user_TwitterUser] := (
	"favorites_count" /. List@@user /. {"favorites_count"->0}
);


(* ::Input:: *)
(*TwitterUserFavoriteCount[user]*)


TwitterUserProtectedQ[user_TwitterUser] := (
	"protected" /. List@@user /. {"protected"->False}
);


(* ::Input:: *)
(*TwitterUserProtectedQ[user]*)


TwitterUserDate[user_TwitterUser] := (
	"created_at" /. List@@user /. {"created_at":>Date[]}
);


(* ::Input:: *)
(*TwitterUserDate[user]*)


TwitterUserInfo[user_TwitterUser] := Row[{
	Style["Name ", Bold],
	TwitterUserName[user],
	"\n",
	Style["Location ", Bold],
	TwitterUserLocation[user],
	"\n",
	Style["Web ", Bold],
	TwitterUserHomePage[user],
	"\n",
	Style["Bio ", Bold],
	TwitterUserDescription[user]
}]


(* ::Input:: *)
(*TwitterUserInfo[user]*)


TwitterUserUI[user_TwitterUser] := 
	Hyperlink[
		Column[{
			Tooltip[
				TwitterUserProfileImage[user],
				TwitterUserInfo[user]
			],
			TwitterUserScreenName[user]
		}, Alignment->Center],
		TwitterUserProfilePage[user]
	];


(* ::Input:: *)
(*TwitterUserUI[user]*)


(* ::Subsection:: *)
(*TwitterStatus*)


Format[status_TwitterStatus, StandardForm] := StringForm[
	"TwitterStatus[<`1`: `2`>]",
	TwitterUserScreenName@TwitterStatusUser[status],
	TwitterStatusText[status]
];


(* ::Input:: *)
(*status*)


TwitterStatus[XMLElement["status", _, elems_List]] := Module[
	{rules, users},
	rules = Cases[elems, XMLElement[tag_String, _, {str_String}] :> Rule[tag, str]];
	users = Cases[elems, XMLElement["user", _, _List]];
	rules = rules /. {
		Rule["id", x_] :> Rule["id", ToExpression[x]],
		Rule["in_reply_to_status_id", x_] :>
			Rule["in_reply_to_status_id", ToExpression[x]],
		Rule["in_reply_to_user_id", x_] :>
			Rule["in_reply_to_user_id", ToExpression[x]],
		Rule["created_at", x_] :>
			Rule["created_at", utcToLocal[HTTPDateList[x]]],
		Rule["truncated", "true"] :> Rule["truncated", True],
		Rule["truncated", "false"] :> Rule["truncated", False],
		Rule["favorited", "true"] :> Rule["favorited", True],
		Rule["favorited", "false"] :> Rule["favorited", False]
	};
	If[Length[users] =!= 0,
		rules = Append[rules, Rule["user", TwitterUser[First@users]]]
	];
	TwitterStatus@@rules
];


TwitterStatus[id_Integer] := Module[
	{xml},
	xml = Import["http://twitter.com/statuses/show/" <>
		ToString[id] <> ".xml"];
	If[xml === $Failed, Return[$Failed]];
	TwitterStatus@First@Cases[xml, XMLElement["status", _, _], \[Infinity]]
];


(* ::Input:: *)
(*TwitterStatus[1533688829]*)


TwitterStatusID[status_TwitterStatus] := (
	"id" /. List@@status /. {"id"->0}
);


(* ::Input:: *)
(*TwitterStatusID[status]*)


TwitterStatusUser[status_TwitterStatus] := (
	"user" /. List@@status /. {"user"->TwitterSessionUser[status]}
)


(* ::Input:: *)
(*TwitterStatusUser[status]*)


TwitterStatusText[status_TwitterStatus] := (
	"text" /. List@@status /. {"text"->""}
);


(* ::Input:: *)
(*TwitterStatusText[status]*)


TwitterStatusDate[status_TwitterStatus] := DateString[
	"created_at" /. List@@status /. {"created_at" :> Date[]}
];


(* ::Input:: *)
(*TwitterStatusDate[status]*)


TwitterStatusPage[status_TwitterStatus] := TwitterStatusPage[
	TwitterUserScreenName[TwitterStatusUser[status]],
	TwitterStatusID[status]
];


TwitterStatusPage[screenName_String, statusID_Integer] := StringJoin[
	"http://twitter.com/",
	screenName,
	"/statuses/",
	ToString[statusID]
];


(* ::Input:: *)
(*TwitterStatusPage[status]*)


TwitterStatusSource[status_TwitterStatus] := Module[
	{source = "source" /. List@@status /. {"source"->"web"}},
	First@StringCases[source, {
		RegularExpression["<a href=\"(.*)\">(.*)</a>"]:>Hyperlink["$2", "$1"],
		x__ :> x
	}]
];


(* ::Input:: *)
(*TwitterStatusSource[status]*)


TwitterStatusReplyID[status_TwitterStatus] := (
	"in_reply_to_status_id" /. List@@status /. {"in_reply_to_status_id"->0}
);


(* ::Input:: *)
(*TwitterStatusReplyID[status]*)


TwitterStatusReplyUserID[status_TwitterStatus] := (
	"in_reply_to_user_id" /. List@@status /. {"in_reply_to_user_id"->0}
);


(* ::Input:: *)
(*TwitterStatusReplyUserID[status]*)


TwitterStatusUI[status_TwitterStatus] := Module[
	{user = TwitterStatusUser[status], replyID, replyUser},
	replyID = TwitterStatusReplyID[status];
	replyUser = If[replyID =!= 0, TwitterUserScreenName[
		TwitterUser[TwitterStatusReplyUserID[status]]],
		""
	];
	Panel@Style[Grid[{{
		TwitterUserProfileImage[user],
		Column[{
			Row[{
				Hyperlink[
					TwitterUserScreenName[user],
					TwitterUserProfilePage[user]
				],
				": ",
				TwitterStatusText[status]
			}],
			Row[{
				Hyperlink[
					TwitterStatusDate[status],
					TwitterStatusPage[status]
				],
				" from ",
				TwitterStatusSource[status],
				Sequence@@If[replyID === 0, {}, {
					" in reply to ",
					Hyperlink[replyUser,
						TwitterStatusPage[replyUser, replyID]]
				}]
			}]
		}, Dividers->{None, 2->Gray}]
	}}], "Text"]
];


(* ::Input:: *)
(*TwitterStatusUI[status]*)


twitterGetXML[url_String, opts___?OptionQ] := Module[
	{twitter, client, closeClient = False, query, u, xml},
	twitter = "Session" /. {opts};
	client = If[Head[twitter] === String,
		"Client" /. {opts},
		TwitterSessionHTTPClient[twitter]
	];
	If[Head[client] === String,
		client = HTTPOpen[];
		closeClient = True;
	];
	
	u = getURLWithQueryString[url, Cases[{opts},
		Rule[lhs:("Page"|"screen_name"|"user_id"), rhs_]:>Rule[ToLowerCase[lhs], rhs]]];
	
	xml = HTTPGet[client, u];
	If[closeClient, HTTPClose[client]];
	If[xml === $Failed, Return[$Failed]];
	
	xml = ImportString[xml, "XML"];
	If[xml === $Failed, Return[$Failed]];
	
	xml
];


twitterGetStatuses[url_String, opts___?OptionQ] := Module[
	{xml = twitterGetXML[url, opts]},
	TwitterStatus /@ Cases[xml[[2]], XMLElement["status", _, _], \[Infinity]]
];


twitterGetUsers[url_String, opts___?OptionQ] := Module[
	{xml = twitterGetXML[url, opts]},
	TwitterUser /@ Cases[xml[[2]], XMLElement["user", _, _], \[Infinity]]
];


(* ::Subsubsection:: *)
(*Set/Delete Status*)


TwitterSetStatus[twitter_TwitterSession, status_] := Module[
	{res, xml},
	res = HTTPPost[TwitterSessionHTTPClient[twitter],
		"https://twitter.com/statuses/update.xml?status=" <>
		hexEncode[ToString[status]]
	];
	If[res === $Failed, Return[$Failed]];
	xml = ImportString[res, "XML"];
	TwitterStatus@First@Cases[xml, XMLElement["status", _, _], \[Infinity]]
];


(* ::Input:: *)
(*newstatus=TwitterSetStatus[twitter,"testing... 1..."]*)


TwitterReply[twitter_TwitterSession, status_TwitterStatus, reply_] :=
	TwitterReply[twitter, TwitterStatusID[status], reply];


TwitterReply[twitter_TwitterSession, status_Integer, reply_] := Module[
	{res},
	res = HTTPPost[TwitterSessionHTTPClient[twitter], StringJoin[
		"https://twitter.com/statuses/update.xml?",
		"in_reply_to_status_id=", ToString[status],
		"&",
		"status=", hexEncode[ToString[reply]]
	]];
	If[res === $Failed, Return[$Failed]];
	xml = ImportString[res, "XML"];
	TwitterStatus@First@Cases[xml, XMLElement["status", _, _], \[Infinity]]
];


TwitterDeleteStatus[twitter_TwitterSession, status_Integer] := Module[
	{res, xml},
	res = HTTPPost[TwitterSessionHTTPClient[twitter],
		"https://twitter.com/statuses/destroy/" <>
		ToString[status] <> ".xml"
	];
	If[res === $Failed, Return[$Failed]];
	xml = ImportString[res, "XML"];
	TwitterStatus@First@Cases[xml, XMLElement["status", _, _], \[Infinity]]
];


TwitterDeleteStatus[twitter_TwitterSession, status_TwitterStatus] :=
	TwitterDeleteStatus[twitter, TwitterStatusID[status]];


(* ::Input:: *)
(*TwitterDeleteStatus[twitter,newstatus]*)


(* ::Subsubsection:: *)
(*Public Timeline*)


(* ::Text:: *)
(*TwitterPublicTimeline[] := Module[*)
(*   	{xml},*)
(*   	xml = Import["http://twitter.com/statuses/public_timeline.xml", "XML"];*)
(*   	If[xml === $Failed, Return[$Failed]];*)
(*   	TwitterStatus /@ Cases[xml[[2]], XMLElement["status", _, _], \[Infinity]]*)
(*   ];*)


TwitterPublicTimeline[opts___?OptionQ] := Module[
	{xml},
	xml = Import[
		getURLWithQueryString[
			"https://twitter.com/statuses/public_timeline.xml",
			{opts}],
		"XML"];
	If[xml === $Failed, Return[$Failed]];
	TwitterStatus /@ Cases[xml[[2]], XMLElement["status", _, _], \[Infinity]]
];


(* ::Input:: *)
(*statuses=TwitterPublicTimeline["Page"->3]*)


(* ::Text:: *)
(*TwitterPublicTimeline[sinceID_Integer] := Module[*)
(*   	{xml},*)
(*   	xml = Import[*)
(*     		"http://twitter.com/statuses/public_timeline.xml?since_id=" <>*)
(*      		ToString[sinceID], "XML"];*)
(*   	If[xml === $Failed, Return[$Failed]];*)
(*   	TwitterStatus /@ Cases[xml[[2]], XMLElement["status", _, _], \[Infinity]]*)
(*   ]; *)


(* ::Text:: *)
(*TwitterPublicTimeline[status_TwitterStatus] :=*)
(*  	TwitterPublicTimeline[TwitterStatusID[status]];*)


(* ::Text:: *)
(*First@TwitterPublicTimeline[status] // InputForm*)


(* ::Subsubsection:: *)
(*Friends Timeline*)


TwitterFriendsTimeline[twitter_TwitterSession, opts___?OptionQ] := 
	twitterGetStatuses[
		"https://twitter.com/statuses/friends_timeline.xml",
		"Session"->twitter,
		opts
	];


(* ::Input:: *)
(*TwitterFriendsTimeline[twitter,"Page"->2]*)


(* ::Subsubsection:: *)
(*User Timeline*)


TwitterUserTimeline[twitter_TwitterSession, opts___?OptionQ] := 
	twitterGetStatuses[
		"https://twitter.com/statuses/user_timeline.xml",
		"Session"->twitter,
		opts
	];


(* ::Input:: *)
(*TwitterUserTimeline[twitter,"Page"->2]*)


TwitterUserTimeline[user:(_Integer|_String), opts___?OptionQ] :=
	twitterGetStatuses[
		"https://twitter.com/statuses/user_timeline/" <>
			ToString[user] <> ".xml",
		opts
	];


(* ::Input:: *)
(*TwitterUserTimeline["melissa_raguet"]*)


(* ::Input:: *)
(*TwitterUserID[TwitterUser["melissa_raguet"]]*)


(* ::Input:: *)
(*TwitterUserTimeline[13140502,"Page"->2]*)


TwitterUserTimeline[user_TwitterUser, opts___?OptionQ] :=
	TwitterUserTimeline[TwitterUserID[user], opts];


(* ::Subsubsection:: *)
(*Replies*)


TwitterReplies[twitter_TwitterSession, opts___?OptionQ] :=
	twitterGetStatuses[
		"https://twitter.com/statuses/replies.xml",
		"Session"->twitter,
		opts
	];


(* ::Input:: *)
(*TwitterReplies[twitter]*)


(* ::Input:: *)
(*TwitterReplies[twitter,"Page"->2]*)


(* ::Input:: *)
(*TwitterReplies[twitter,"Page"->10]*)


(* ::Subsubsection:: *)
(*Friends*)


TwitterFriends[twitter_TwitterSession, opts___?OptionQ] :=
	twitterGetUsers[
		"https://twitter.com/statuses/friends.xml",
		"Session"->twitter,
		opts
	];


(* ::Input:: *)
(*TwitterFriends[twitter]//Length*)


TwitterFriends[twitter_TwitterSession, user:(_Integer|_String), opts___?OptionQ] :=
	TwitterFriends[twitter,
		If[Head[user]===String, "screen_name", "user_id"]->ToString[user],
		opts
	];


(* ::Input:: *)
(*TwitterFriends[twitter,"melissa_raguet"]//Length*)


(* ::Input:: *)
(*TwitterFriends[twitter,12920162]//Length*)


TwitterFriends[twitter_TwitterSession, user_TwitterUser, opts___?OptionQ] :=
	TwitterFriends[twitter, TwitterUserID[user], opts];


(* ::Input:: *)
(*TwitterFriends[twitter,user]//Length*)


twitterAllUsers[twitter_TwitterSession, url_String, user:(_Integer|_String), opts___?OptionQ] := Module[
	{xml, pageResults, allResults = {}, page = 0},
	Monitor[
		While[True,
			pageResults = twitterGetUsers[
				url,
				"Session"->twitter,
				If[Head[user]===String, "screen_name", "user_id"]->ToString[user],
				"Page"->(++page),
				opts
			];
			If[Length[pageResults] === 0, Break[]];
			allResults = Join[allResults, pageResults];
		],
		StringJoin["Importing page ", ToString[page],
			" (", ToString[Length[allResults]], " results)"]
	];
	allResults
];


(* ::Input:: *)
(*users=Twitter`Private`twitterAllUsers[twitter,"https://twitter.com/statuses/friends.xml","WolframResearch"];*)


(* ::Input:: *)
(*Length[users]*)


(* ::Input:: *)
(*users=Twitter`Private`twitterAllUsers[twitter,"https://twitter.com/statuses/followers.xml","WolframResearch"];*)
(*Length[users]*)


TwitterAllFriends[twitter_TwitterSession, opts___?OptionQ] :=
	TwitterAllFriends[twitter, TwitterSessionScreenName[twitter], opts];


TwitterAllFriends[twitter_TwitterSession, user:(_Integer|_String), opts___?OptionQ] :=
	twitterAllUsers[twitter, "https://twitter.com/statuses/friends.xml", user, opts];


TwitterAllFriends[twitter_TwitterSession, user_TwitterUser, opts___?OptionQ] :=
	TwitterAllFriends[twitter, TwitterUserID[user], opts];


(* ::Input:: *)
(*TwitterAllFriends[twitter]//Length*)


(* ::Input:: *)
(*TwitterAllFriends[twitter,"melissa_raguet"]//Length*)


(* ::Subsubsection:: *)
(*Followers*)


TwitterFollowers[twitter_TwitterSession, opts___?OptionQ] :=
	twitterGetUsers[
		"https://twitter.com/statuses/followers.xml",
		"Session"->twitter,
		opts
	];


(* ::Input:: *)
(*TwitterFollowers[twitter]//Length*)


TwitterFollowers[twitter_TwitterSession, user:(_Integer|_String), opts___?OptionQ] :=
	TwitterFollowers[twitter,
		If[Head[user]===String, "screen_name", "user_id"]->ToString[user],
		opts
	];


(* ::Input:: *)
(*TwitterFollowers[twitter,"melissa_raguet"]//Length*)


TwitterFollowers[twitter_TwitterSession, user_TwitterUser, opts___?OptionQ] :=
	TwitterFollowers[twitter, TwitterUserID[user], opts];


(* ::Input:: *)
(*TwitterFollowers[twitter,user]//Length*)


(* ::Input:: *)
(*TwitterFollowers[twitter,12920162]//Length*)


TwitterAllFollowers[twitter_TwitterSession, opts___?OptionQ] :=
	TwitterAllFollowers[twitter, TwitterSessionScreenName[twitter], opts];


TwitterAllFollowers[twitter_TwitterSession, user:(_Integer|_String), opts___?OptionQ] :=
	twitterAllUsers[twitter, "https://twitter.com/statuses/followers.xml", user, opts];


TwitterAllFollowers[twitter_TwitterSession, user_TwitterUser, opts___?OptionQ] :=
	TwitterAllFollowers[twitter, TwitterUserID[user], opts];


(* ::Input:: *)
(*TwitterAllFollowers[twitter]//Length*)


(* ::Input:: *)
(*TwitterAllFollowers[twitter,"melissa_raguet"]//Length*)


(* ::Section:: *)
(*Finalization*)


End[]; (* `Private` *)


EndPackage[]; (* Twitter` *)
